/********************************************************************
      
      mod_anti_tamper (apache 2.x), http://www.wisec.it
      
     Copyright (c) 2005 Stefano Di Paola
     Author: Stefano Di Paola <stefano.dipaola@wisec.it>



This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
     
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Note:
     The html parser was heavily taken from mod_proxy_html.c
	 Copyright (c) 2003-4, WebThing Ltd
	 Author: Nick Kew <nick@webthing.com>

      
*********************************************************************/



/* End of Notices */




/* 	GO_FASTER

	You can #define GO_FASTER to disable informational logging.
	This disables the ProxyHTMLLogVerbose option altogether.

	Default is to leave it undefined, and enable verbose logging
	as a configuration option.  Binaries are supplied with verbose
	logging enabled.
*/
//#define GO_FASTER

#ifdef GO_FASTER
#define VERBOSE(x)
#else
#define VERBOSE(x) if ( verbose ) x
#endif

#define VERSION_STRING "anti_tamper/0.5"

#include <ctype.h>

/* libxml */
#include <libxml/HTMLparser.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <dlfcn.h>
/* apache */
#include <http_protocol.h>
#include <http_config.h>
#include <http_log.h>
#include <apr_strings.h>
#include <apr_sha1.h>
#include "http_request.h"

module AP_MODULE_DECLARE_DATA anti_tamper_module;

#define M_HTML		0x01
#define M_EVENTS	0x02
#define M_CDATA		0x04
#define M_REGEX		0x08
#define M_ATSTART	0x10
#define M_ATEND		0x20
#define M_LAST		0x40

#define ATPASSWDLEN 40
#define KEY_PASS "123"
#define QUERY_SEPARATOR "&"
#define COOKIE_PREFIX ":"


#define DEFAULT_AUTHVAR_PREFIX "_auth="
#define DEFAULT_HMACRESP_LEN 40
#define DEFAULT_DONTMODIFYVAR_PREFIX "dontmodify"
#define DEFAULT_REDIRECTION "/"

#define SECRET_PATH "/etc/httpd/conf/anti_tamper_secret"

typedef struct
{
  unsigned int start;
  unsigned int end;
} meta;
typedef struct urlmap
{
  struct urlmap *next;
  unsigned int flags;
  union
  {
    const char *c;
    regex_t *r;
  } from;
  const char *to;
} urlmap;


typedef struct
{
  urlmap *map;
  const char *doctype;
  const char *etag;
  unsigned int flags;
  int wants_url_anti_tamper;
  int wants_cookie_anti_tamper;
  apr_array_header_t *exclude_args;
  char *hmacprefix;
  int url_hmacresponselen;
  int cookie_hmacresponselen;
  int url_useip;
  int cookie_useip;
  int extfix;
  int metafix;
  int strip_comments;
  char *secret_path;
  char atpasswd[ATPASSWDLEN];
  char *(*url_funchandle) (request_rec *, char *, int, const unsigned char *,
		       int);
  char *(*cookie_funchandle) (request_rec *, char *, int, const unsigned char *,
		       int);
  char *redir_url;
  char *dontmodifyprefix;
#ifndef GO_FASTER
  int verbose;
#endif
  size_t bufsz;
} anti_tamper_conf;

static ap_filter_rec_t *global_at_filter_out;


typedef struct
{
  htmlSAXHandlerPtr sax;
  ap_filter_t *f;
  anti_tamper_conf *cfg;
  htmlParserCtxtPtr parser;
  apr_bucket_brigade *bb;
  apr_table_t *cookie_table;
  char *buf;
  size_t offset;
  size_t avail;
  int got_headers;
} saxctxt;

static int
is_empty_elt (const char *name)
{
  const char **p;
  static const char *empty_elts[] = {
    "br",
    "link",
    "img",
    "hr",
    "input",
    "meta",
    "base",
    "area",
    "param",
    "col",
    "frame",
    "isindex",
    "basefont",
    NULL
  };
  for (p = empty_elts; *p; ++p)
    if (!strcmp (*p, name))
      return 1;
  return 0;
}

typedef struct
{
  const char *name;
  const char **attrs;
} elt_t;



#define NORM_LC 0x1
#define NORM_MSSLASH 0x2
#define NORM_RESET 0x4


typedef enum
{ ATTR_IGNORE, ATTR_URI, ATTR_EVENT } rewrite_t;


char *is_cookie_ok (const char *cookie, request_rec * r, anti_tamper_conf * dcfg);


static char *
get_hmac_string (request_rec * r, char *msg, int msglen,
		 const unsigned char *key, int key_len)
{
  apr_sha1_ctx_t ctx;
  unsigned char digest[APR_SHA1_DIGESTSIZE];
  unsigned char k_ipad[65], k_opad[65];
  unsigned char kt[APR_SHA1_DIGESTSIZE];
  unsigned char *k = (unsigned char *) key;
  char hex_digest[APR_SHA1_DIGESTSIZE * 2], *hd;
  const char hex[] = "0123456789abcdef";
  int i;
  apr_pool_t *p = r->pool;

  /* if key is longer than 64 bytes reset it to key=SHA1(key) */
  if (key_len > 64)
    {
      k = kt;
      key_len = APR_SHA1_DIGESTSIZE;
    }

  /* HMAC_SHA1 transform */
  /* start out by string key in pads */
  memset ((void *) k_ipad, 0, sizeof (k_ipad));
  memset ((void *) k_opad, 0, sizeof (k_opad));
  memmove (k_ipad, k, key_len);
  memmove (k_opad, k, key_len);

  /* XOR key with ipad and opad values */
  for (i = 0; i < 64; i++)
    {
      k_ipad[i] ^= 0x36;
      k_opad[i] ^= 0x5c;
    }
  /* perform inner SHA-1 */
  apr_sha1_init (&ctx);
  apr_sha1_update_binary (&ctx, k_ipad, 64);
  apr_sha1_update_binary (&ctx, (const unsigned char *) msg, msglen);
  apr_sha1_final (digest, &ctx);
  /* perform outer SHA-1 */
  apr_sha1_init (&ctx);
  apr_sha1_update_binary (&ctx, k_opad, 64);
  apr_sha1_update_binary (&ctx, digest, sizeof (digest));
  apr_sha1_final (digest, &ctx);

  hd = hex_digest;
  for (i = 0; i < sizeof (digest); i++)
    {
      *hd++ = hex[digest[i] >> 4];
      *hd++ = hex[digest[i] & 0xF];
    }
  *hd = '\0';

  return apr_pstrdup (p, hex_digest);
}

static unsigned char *
base_uri (char *buf, request_rec * r, unsigned char *append_auth)
{
  unsigned char *dirname;
  int no_absolute = 0, is_from_us = 0;

  if (*(buf) == '/')
    return append_auth;



  if (*(buf) != '/' && (no_absolute = strncasecmp (buf, "http://", 7)))
    {
      char *lastsep = strrchr (r->uri, '/');

      if (lastsep)
	dirname = apr_pstrndup (r->pool, r->uri, lastsep + 1 - (r->uri));
      else
	{
	  dirname =
	    (unsigned char *) apr_pcalloc (r->pool,
					   sizeof (unsigned char) * 2);
	  dirname[0] = '/';
	  dirname[1] = '\0';
	}
      append_auth = apr_pstrcat (r->pool, dirname, append_auth, NULL);
      return append_auth;
    }
  // f->r->server->server_hostname,apr_table_get(f->r->headers_in, "Host"), f->r->connection->local_ip
  is_from_us = (strstr (buf + 7, r->server->server_hostname)
		|| strstr (buf + 7, r->connection->local_ip));
  if (!no_absolute && is_from_us)
    {
      char *firstsep = NULL;
      char *querypos = NULL;

      querypos = strchr (buf + 7, '?');

      firstsep = strchr (buf + 7, '/');


      if (firstsep)
	append_auth = apr_pstrdup (r->pool, firstsep);

      if (firstsep > querypos || firstsep == NULL)
	append_auth = apr_pstrcat (r->pool,  DEFAULT_REDIRECTION , querypos, NULL);


    }
  else
    {
      return NULL;
    }

  return append_auth;
}

static void
normalise (unsigned int flags, char *str)
{
  xmlChar *p;
  if (flags & NORM_LC)
    for (p = str; *p; ++p)
      if (isupper (*p))
	*p = tolower (*p);

  if (flags & NORM_MSSLASH)
    for (p = strchr (str, '\\'); p; p = strchr (p + 1, '\\'))
      *p = '/';

}

#define FLUSH ap_fwrite(ctx->f->next, ctx->bb, (chars+begin), (i-begin)) ; begin = i+1
static void
pcharacters (void *ctxt, const xmlChar * chars, int length)
{
  saxctxt *ctx = (saxctxt *) ctxt;
  int i;
  int begin;
  for (begin = i = 0; i < length; i++)
    {
      switch (chars[i])
	{
	case '&':
	  FLUSH;
	  ap_fputs (ctx->f->next, ctx->bb, "&amp;");
	  break;
	case '<':
	  FLUSH;
	  ap_fputs (ctx->f->next, ctx->bb, "&lt;");
	  break;
	case '>':
	  FLUSH;
	  ap_fputs (ctx->f->next, ctx->bb, "&gt;");
	  break;
	case '"':
	  FLUSH;
	  ap_fputs (ctx->f->next, ctx->bb, "&quot;");
	  break;
	default:
	  break;
	}
    }
  FLUSH;
}

static void
preserve (saxctxt * ctx, const size_t len)
{
  char *newbuf;
  if (len <= (ctx->avail - ctx->offset))
    return;
  else
    while (len > (ctx->avail - ctx->offset))
      ctx->avail += ctx->cfg->bufsz;

  newbuf = realloc (ctx->buf, ctx->avail);
  if (newbuf != ctx->buf)
    {
      if (ctx->buf)
	apr_pool_cleanup_kill (ctx->f->r->pool, ctx->buf, (void *) free);
      apr_pool_cleanup_register (ctx->f->r->pool, newbuf,
				 (void *) free, apr_pool_cleanup_null);
      ctx->buf = newbuf;
    }
}

static void
pappend (saxctxt * ctx, const char *buf, const size_t len)
{
  preserve (ctx, len);
  memcpy (ctx->buf + ctx->offset, buf, len);
  ctx->offset += len;
}

static void
dump_content (saxctxt * ctx)
{
  urlmap *m;
  char *found;
  size_t s_from, s_to;
  size_t match;
  char c = 0;
  int nmatch;
  regmatch_t pmatch[10];
  char *subs;
  size_t len, offs;
#ifndef GO_FASTER
  int verbose = ctx->cfg->verbose;
#endif

  pappend (ctx, &c, 1);		/* append null byte */
  /* parse the text for URLs */

  for (m = ctx->cfg->map; m; m = m->next)
    {
      if (!(m->flags & M_CDATA))
	continue;
      if (m->flags & M_REGEX)
	{
	  nmatch = 10;
	  offs = 0;
	  while (!ap_regexec (m->from.r, ctx->buf + offs, nmatch, pmatch, 0))
	    {
	      match = pmatch[0].rm_so;
	      s_from = pmatch[0].rm_eo - match;
	      subs = ap_pregsub (ctx->f->r->pool, m->to, ctx->buf + offs,
				 nmatch, pmatch);
	      s_to = strlen (subs);
	      len = strlen (ctx->buf);
	      offs += match;
	      VERBOSE (
			{
			const char *f = apr_pstrndup (ctx->f->r->pool,
						      ctx->buf + offs,
						      s_from);
			ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, ctx->f->r,
				       "C/RX: match at %s, substituting %s",
				       f, subs);}) if (s_to > s_from)
		{
		  preserve (ctx, s_to - s_from);
		  memmove (ctx->buf + offs + s_to, ctx->buf + offs + s_from,
			   len + 1 - s_from - offs);
		  memcpy (ctx->buf + offs, subs, s_to);
		}
	      else
		{
		  memcpy (ctx->buf + offs, subs, s_to);
		  memmove (ctx->buf + offs + s_to, ctx->buf + offs + s_from,
			   len + 1 - s_from - offs);
		}
	      offs += s_to;
	    }
	}
      else
	{
	  s_from = strlen (m->from.c);
	  s_to = strlen (m->to);
	  for (found = strstr (ctx->buf, m->from.c); found;
	       found = strstr (ctx->buf + match + s_to, m->from.c))
	    {
	      match = found - ctx->buf;
	      if ((m->flags & M_ATSTART) && (match != 0))
		break;
	      len = strlen (ctx->buf);
	      if ((m->flags & M_ATEND) && (match < (len - s_from)))
		continue;
	      VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, ctx->f->r,
				      "C: matched %s, substituting %s",
				      m->from.c, m->to));
	      if (s_to > s_from)
		{
		  preserve (ctx, s_to - s_from);
		  memmove (ctx->buf + match + s_to, ctx->buf + match + s_from,
			   len + 1 - s_from - match);
		  memcpy (ctx->buf + match, m->to, s_to);
		}
	      else
		{
		  memcpy (ctx->buf + match, m->to, s_to);
		  memmove (ctx->buf + match + s_to, ctx->buf + match + s_from,
			   len + 1 - s_from - match);
		}
	    }
	}
    }
  ap_fputs (ctx->f->next, ctx->bb, ctx->buf);
}


static void
pcdata (void *ctxt, const xmlChar * chars, int length)
{
  saxctxt *ctx = (saxctxt *) ctxt;
  if (ctx->cfg->extfix)
    {
      pappend (ctx, chars, length);
    }
  else
    {
      ap_fwrite (ctx->f->next, ctx->bb, chars, length);
    }
}



static void
pcomment (void *ctxt, const xmlChar * chars)
{
  saxctxt *ctx = (saxctxt *) ctxt;
  if (ctx->cfg->strip_comments)
    return;

  if (ctx->cfg->extfix)
    {
      pappend (ctx, "<!--", 4);
      pappend (ctx, chars, strlen (chars));
      pappend (ctx, "-->", 3);
    }
  else
    {
      ap_fputstrs (ctx->f->next, ctx->bb, "<!--", chars, "-->", NULL);
    }
}



static void
pendElement (void *ctxt, const xmlChar * name)
{
  saxctxt *ctx = (saxctxt *) ctxt;
  if (ctx->offset > 0)
    {
      dump_content (ctx);
      ctx->offset = 0;		/* having dumped it, we can re-use the memory */
    }
  if (!is_empty_elt (name))
    ap_fprintf (ctx->f->next, ctx->bb, "</%s>", name);
}


static void
pstartElement (void *ctxt, const xmlChar * name, const xmlChar ** attrs)
{

  int num_match;
  size_t offs, len;
  char *subs;
  rewrite_t is_uri;
  const char **linkattrs;
  const xmlChar **a;
  const elt_t *elt;
  const char **linkattr;
  urlmap *m;
  size_t s_to, s_from, match;
  char *found;
  saxctxt *ctx = (saxctxt *) ctxt;
  size_t nmatch;
  regmatch_t pmatch[10];
#ifndef GO_FASTER
  int verbose = ctx->cfg->verbose;
#endif

  static const char *href[] = { "href", NULL };
  static const char *cite[] = { "cite", NULL };
  static const char *action[] = { "action", NULL };
  static const char *imgattr[] = { "src", "longdesc", "usemap", NULL };
  static const char *inputattr[] = { "src", "usemap", NULL };
  static const char *scriptattr[] = { "src", "for", NULL };
  static const char *frameattr[] = { "src", "longdesc", NULL };
  static const char *objattr[] =
    { "classid", "codebase", "data", "usemap", NULL };
  static const char *profile[] = { "profile", NULL };
  static const char *background[] = { "background", NULL };
  static const char *codebase[] = { "codebase", NULL };

  static const elt_t linked_elts[] = {
    {"a", href},
    {"img", imgattr},
    {"form", action},
    {"link", href},
    {"script", scriptattr},
    {"base", href},
    {"area", href},
    {"input", inputattr},
    {"frame", frameattr},
    {"iframe", frameattr},
    {"object", objattr},
    {"q", cite},
    {"blockquote", cite},
    {"ins", cite},
    {"del", cite},
    {"head", profile},
    {"body", background},
    {"applet", codebase},
    {NULL, NULL}
  };
  static const char *events[] = {
    "onclick",
    "ondblclick",
    "onmousedown",
    "onmouseup",
    "onmouseover",
    "onmousemove",
    "onmouseout",
    "onkeypress",
    "onkeydown",
    "onkeyup",
    "onfocus",
    "onblur",
    "onload",
    "onunload",
    "onsubmit",
    "onreset",
    "onselect",
    "onchange",
    NULL
  };

  ap_fputc (ctx->f->next, ctx->bb, '<');
  ap_fputs (ctx->f->next, ctx->bb, name);

  if (attrs)
    {
      linkattrs = 0;
      for (elt = linked_elts; elt->name != NULL; ++elt)
	if (!strcmp (elt->name, name))
	  {
	    linkattrs = elt->attrs;
	    break;
	  }
      for (a = attrs; *a; a += 2)
	{
	  ctx->offset = 0;
	  if (a[1])
	    {
	      pappend (ctx, a[1], strlen (a[1]) + 1);
	      is_uri = ATTR_IGNORE;
	      if (linkattrs)
		{
		  for (linkattr = linkattrs; *linkattr; ++linkattr)
		    {
		      if (!strcmp (*linkattr, *a))
			{
			  is_uri = ATTR_URI;
			  break;
			}
		    }
		}
	      if ((is_uri == ATTR_IGNORE) && ctx->cfg->extfix)
		{
		  for (linkattr = events; *linkattr; ++linkattr)
		    {
		      if (!strcmp (*linkattr, *a))
			{
			  is_uri = ATTR_EVENT;
			  break;
			}
		    }
		}
	      switch (is_uri)
		{
		case ATTR_URI:
		  num_match = 0;
		  if (ctx->cfg->wants_url_anti_tamper
		      && !strstr (ctx->buf, ctx->cfg->dontmodifyprefix))
		    {
		      if (strstr (ctx->buf, "?"))
			{
			  unsigned char   *append_auth, *anchor_name,*modified_url;
			  unsigned char *anchorsep = 0;
			  append_auth =
			    apr_pstrdup (ctx->f->r->pool, ctx->buf);
			  anchor_name = 0;
 			  // Base URI
			  append_auth =
			    base_uri (ctx->buf, ctx->f->r, append_auth);
 
			  if (append_auth)
			    {

			      // Remove # anchor for HMAC
			      if ((anchorsep = strrchr (append_auth, '#')))
				{
				  unsigned char *splitted_anchor =
				    apr_pstrndup (ctx->f->r->pool,
						  append_auth,
						  anchorsep - append_auth);
				  anchor_name =
				    apr_pstrndup (ctx->f->r->pool, anchorsep,
						  strlen (anchorsep));
				  append_auth =
				    apr_pstrdup (ctx->f->r->pool,
						 splitted_anchor);

				}
				  modified_url = apr_pstrndup (ctx->f->r->pool,append_auth  , strlen(append_auth) );
 				 if(  ctx->cfg->url_useip)
				    modified_url =apr_pstrcat(ctx->f->r->pool,modified_url,ctx->f->r->connection->remote_ip,NULL);
 
			       unsigned char *hmac =
				 ctx->cfg->url_funchandle (ctx->f->r,modified_url  ,
					  	       strlen (modified_url ),
						       ctx->cfg->atpasswd, ATPASSWDLEN);



			      append_auth =
				apr_pstrcat (ctx->f->r->pool, append_auth,
					     QUERY_SEPARATOR,
					     ctx->cfg->hmacprefix, hmac,
					     NULL);

			      if (anchor_name)
				append_auth =
				  apr_pstrcat (ctx->f->r->pool, append_auth,
					       anchor_name, NULL);

			      s_to = strlen (append_auth);
			      len = strlen (ctx->buf);

			      s_from = len;
			      if (s_to > s_from)
				{
				  preserve (ctx, s_to - s_from);

				  memmove (ctx->buf + s_to, ctx->buf + s_from,
					   len + 1 - s_from);

				  memcpy (ctx->buf, append_auth, s_to);
				}


			    }
			}
		    }
		  for (m = ctx->cfg->map; m; m = m->next)
		    {
		      if (!(m->flags & M_HTML))
			continue;

		      if (m->flags & M_REGEX)
			{
			  nmatch = 10;
			  if (!ap_regexec
			      (m->from.r, ctx->buf, nmatch, pmatch, 0))
			    {
			      ++num_match;
			      offs = match = pmatch[0].rm_so;
			      s_from = pmatch[0].rm_eo - match;
			      subs =
				ap_pregsub (ctx->f->r->pool, m->to,
					    ctx->buf + offs, nmatch, pmatch);
			      VERBOSE (
					{
					const char *f =
					apr_pstrndup (ctx->f->r->pool,
						      ctx->buf + offs,
						      s_from);
					ap_log_rerror (APLOG_MARK, APLOG_INFO,
						       0, ctx->f->r,
						       "H/RX: match at %s, substituting %s",
						       f, subs);}) s_to =
				strlen (subs);
			      len = strlen (ctx->buf);
			      if (s_to > s_from)
				{
				  preserve (ctx, s_to - s_from);
				  memmove (ctx->buf + offs + s_to,
					   ctx->buf + offs + s_from,
					   len + 1 - s_from - offs);
				  memcpy (ctx->buf + offs, subs, s_to);
				}
			      else
				{
				  memcpy (ctx->buf + offs, subs, s_to);
				  memmove (ctx->buf + offs + s_to,
					   ctx->buf + offs + s_from,
					   len + 1 - s_from - offs);
				}
			    }
			}
		      else
			{
			  s_from = strlen (m->from.c);
			  if (!strncasecmp (ctx->buf, m->from.c, s_from))
			    {
			      ++num_match;
			      s_to = strlen (m->to);
			      len = strlen (ctx->buf);
			      VERBOSE (ap_log_rerror
				       (APLOG_MARK, APLOG_INFO, 0, ctx->f->r,
					"H: matched %s, substituting %s",
					m->from.c, m->to));
			      if (s_to > s_from)
				{
				  preserve (ctx, s_to - s_from);

				  memmove (ctx->buf + s_to, ctx->buf + s_from,
					   len + 1 - s_from);

				  memcpy (ctx->buf, m->to, s_to);
				}
			      else
				{	/* it fits in the existing space */

				  memcpy (ctx->buf, m->to, s_to);

				  memmove (ctx->buf + s_to, ctx->buf + s_from,
					   len + 1 - s_from);
				}
			      break;
			    }
			}
		      if (num_match > 0)	/* URIs only want one match */
			break;
		    }
		  break;
		case ATTR_EVENT:
		  for (m = ctx->cfg->map; m; m = m->next)
		    {
		      num_match = 0;	/* reset here since we're working per-rule */
		      if (!(m->flags & M_EVENTS))
			continue;
		      if (m->flags & M_REGEX)
			{
			  nmatch = 10;
			  offs = 0;
			  while (!ap_regexec (m->from.r, ctx->buf + offs,
					      nmatch, pmatch, 0))
			    {
			      match = pmatch[0].rm_so;
			      s_from = pmatch[0].rm_eo - match;
			      subs =
				ap_pregsub (ctx->f->r->pool, m->to,
					    ctx->buf + offs, nmatch, pmatch);
			      VERBOSE (
					{
					const char *f =
					apr_pstrndup (ctx->f->r->pool,
						      ctx->buf + offs,
						      s_from);
					ap_log_rerror (APLOG_MARK, APLOG_INFO,
						       0, ctx->f->r,
						       "E/RX: match at %s, substituting %s",
						       f, subs);}) s_to =
				strlen (subs);
			      offs += match;
			      len = strlen (ctx->buf);
			      if (s_to > s_from)
				{
				  preserve (ctx, s_to - s_from);
				  memmove (ctx->buf + offs + s_to,
					   ctx->buf + offs + s_from,
					   len + 1 - s_from - offs);
				  memcpy (ctx->buf + offs, subs, s_to);
				}
			      else
				{
				  memcpy (ctx->buf + offs, subs, s_to);
				  memmove (ctx->buf + offs + s_to,
					   ctx->buf + offs + s_from,
					   len + 1 - s_from - offs);
				}
			      offs += s_to;
			      ++num_match;
			    }
			}
		      else
			{
			  found = strstr (ctx->buf, m->from.c);
			  if ((m->flags & M_ATSTART) && (found != ctx->buf))
			    continue;
			  while (found)
			    {
			      s_from = strlen (m->from.c);
			      s_to = strlen (m->to);
			      match = found - ctx->buf;
			      if ((s_from < strlen (found))
				  && (m->flags & M_ATEND))
				{
				  found =
				    strstr (ctx->buf + match + s_from,
					    m->from.c);
				  continue;
				}
			      else
				{
				  found =
				    strstr (ctx->buf + match + s_to,
					    m->from.c);
				}
			      VERBOSE (ap_log_rerror
				       (APLOG_MARK, APLOG_INFO, 0, ctx->f->r,
					"E: matched %s, substituting %s",
					m->from.c, m->to));
			      len = strlen (ctx->buf);
			      if (s_to > s_from)
				{
				  preserve (ctx, s_to - s_from);
				  memmove (ctx->buf + match + s_to,
					   ctx->buf + match + s_from,
					   len + 1 - s_from - match);
				  memcpy (ctx->buf + match, m->to, s_to);
				}
			      else
				{
				  memcpy (ctx->buf + match, m->to, s_to);
				  memmove (ctx->buf + match + s_to,
					   ctx->buf + match + s_from,
					   len + 1 - s_from - match);
				}
			      ++num_match;
			    }
			}
		      if (num_match && (m->flags & M_LAST))
			break;
		    }
		  break;
		case ATTR_IGNORE:
		  break;
		}
	    }
	  if (!a[1])
	    ap_fputstrs (ctx->f->next, ctx->bb, " ", a[0], NULL);
	  else
	    {

	      if (ctx->cfg->flags != 0)
		normalise (ctx->cfg->flags, ctx->buf);

	      /* write the attribute, using pcharacters to html-escape
	         anything that needs it in the value.
	       */
	      ap_fputstrs (ctx->f->next, ctx->bb, " ", a[0], "=\"", NULL);
	      pcharacters (ctx, ctx->buf, strlen (ctx->buf));
	      ap_fputc (ctx->f->next, ctx->bb, '"');
	    }
	}
    }
  ctx->offset = 0;
  if (is_empty_elt (name))
    ap_fputs (ctx->f->next, ctx->bb, ctx->cfg->etag);
  else
    ap_fputc (ctx->f->next, ctx->bb, '>');
}


static htmlSAXHandlerPtr
setupSAX (apr_pool_t * pool)
{
  htmlSAXHandlerPtr sax = apr_pcalloc (pool, sizeof (htmlSAXHandler));
  sax->startDocument = NULL;
  sax->endDocument = NULL;
  sax->startElement = pstartElement;
  sax->endElement = pendElement;
  sax->characters = pcharacters;
  sax->comment = pcomment;
  sax->cdataBlock = pcdata;
  return sax;
}

static regex_t *seek_meta_ctype;
static regex_t *seek_charset;
static regex_t *seek_meta;

static void
anti_tamper_child_init (apr_pool_t * pool, server_rec * s)
{
  seek_meta_ctype = ap_pregcomp (pool,
				 "(<meta[^>]*http-equiv[ \t\r\n='\"]*content-type[^>]*>)",
				 REG_EXTENDED | REG_ICASE);
  seek_charset = ap_pregcomp (pool, "charset=([A-Za-z0-9_-]+)",
			      REG_EXTENDED | REG_ICASE);
  seek_meta = ap_pregcomp (pool, "<meta[^>]*(http-equiv)[^>]*>",
			   REG_EXTENDED | REG_ICASE);
}


static xmlCharEncoding
sniff_encoding (request_rec * r, const char *cbuf, size_t bytes
#ifndef GO_FASTER
		, int verbose
#endif
  )
{
  xmlCharEncoding ret;
  char *encoding = NULL;
  char *p;
  regmatch_t match[2];
  unsigned char *buf = (unsigned char *) cbuf;

  VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, r,
			  "Content-Type is %s", r->content_type));

/* If we've got it in the HTTP headers, there's nothing to do */
  if (r->content_type &&
      (p = ap_strcasestr (r->content_type, "charset="), p > 0))
    {
      p += 8;
      if (encoding = apr_pstrndup (r->pool, p, strcspn (p, " ;")), encoding)
	{
	  if (ret = xmlParseCharEncoding (encoding),
	      ret != XML_CHAR_ENCODING_ERROR)
	    {
	      VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, r,
				      "Got charset %s from HTTP headers",
				      encoding));
	      return ret;
	    }
	  else
	    {
	      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
			     "Unsupported charset %s in HTTP headers",
			     encoding);
	      encoding = NULL;
	    }
	}
    }

/* to sniff, first we look for BOM */
  if (ret = xmlDetectCharEncoding (buf, bytes), ret != XML_CHAR_ENCODING_NONE)
    {
      VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, r,
			      "Got charset from XML rules."));
      return ret;
    }

/* If none of the above, look for a META-thingey */
  encoding = NULL;
  if (ap_regexec (seek_meta_ctype, buf, 1, match, 0) == 0)
    {
      p = apr_pstrndup (r->pool, buf + match[0].rm_so,
			match[0].rm_eo - match[0].rm_so);
      if (ap_regexec (seek_charset, p, 2, match, 0) == 0)
	encoding = apr_pstrndup (r->pool, p + match[1].rm_so,
				 match[1].rm_eo - match[1].rm_so);
    }

/* either it's set to something we found or it's still the default */
  if (encoding){
    if (ret = xmlParseCharEncoding (encoding), ret != XML_CHAR_ENCODING_ERROR)
      {
	VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, r,
				"Got charset %s from HTML META", encoding));
	return ret;
      }
    else
      {
	ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		       "Unsupported charset %s in HTML META", encoding);
      }
}
/* the old HTTP default is a last resort */
  ap_log_rerror (APLOG_MARK, APLOG_WARNING, 0, r,
		 "No usable charset information: using old HTTP default LATIN1");
  return XML_CHAR_ENCODING_8859_1;
}


static meta *
metafix (request_rec * r, const char *buf	/*, size_t bytes */
#ifndef GO_FASTER
	 , int verbose
#endif
  )
{
  meta *ret = NULL;
  size_t offs = 0;
  const char *p;
  const char *q;
  char *header;
  char *content;
  regmatch_t pmatch[2];
  char delim;

  while (!ap_regexec (seek_meta, buf + offs, 2, pmatch, 0))
    {
      header = NULL;
      content = NULL;
      p = buf + offs + pmatch[1].rm_eo;
      while (!isalpha (*++p));
      for (q = p; isalnum (*q) || (*q == '-'); ++q);
      header = apr_pstrndup (r->pool, p, q - p);
      if (strncasecmp (header, "Content-", 8))
	{
/* find content=... string */
	  for (p = strstr ((char *)(buf + offs + pmatch[0].rm_so), "content"); *p;)
	    {
	      p += 7;
	      while (*p && isspace (*p))
		++p;
	      if (*p != '=')
		continue;
	      while (*p && isspace (*++p));
	      if ((*p == '\'') || (*p == '"'))
		{
		  delim = *p++;
		  for (q = p; *q != delim; ++q);
		}
	      else
		{
		  for (q = p; *q && !isspace (*q) && (*q != '>'); ++q);
		}
	      content = apr_pstrndup (r->pool, p, q - p);
	      break;
	    }
	}
      else if (!strncasecmp (header, "Content-Type", 12))
	{
	  ret = apr_palloc (r->pool, sizeof (meta));
	  ret->start = pmatch[0].rm_so;
	  ret->end = pmatch[0].rm_eo;
	}
      if (header && content)
	{
	  VERBOSE (ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, r,
				  "Adding header [%s: %s] from HTML META",
				  header, content));
	  apr_table_setn (r->headers_out, header, content);
	}
      offs += pmatch[0].rm_eo;
    }
  return ret;
}


static int
anti_tamper_filter_init (ap_filter_t * f)
{
  const char *env;
  saxctxt *fctx;
  char *locatio;

#if 0
/* remove content-length filter */
  ap_filter_rec_t *clf = ap_get_output_filter_handle ("CONTENT_LENGTH");
  ap_filter_t *ff = f->next;

  do
    {
      ap_filter_t *fnext = ff->next;
      if (ff->frec == clf)
	ap_remove_output_filter (ff);
      ff = fnext;
    }
  while (ff);
#endif

  fctx = f->ctx = apr_pcalloc (f->r->pool, sizeof (saxctxt));
  fctx->sax = setupSAX (f->r->pool);
  fctx->f = f;
  fctx->bb = apr_brigade_create (f->r->pool, f->r->connection->bucket_alloc);
  fctx->cfg = ap_get_module_config (f->r->per_dir_config, &anti_tamper_module);
  fctx->got_headers = 0;
//  f->r->server->server_hostname,apr_table_get(f->r->headers_in, "Host"), f->r->connection->local_ip
  if ((locatio = (char *) apr_table_get (f->r->headers_out, "Location"))
      && strchr (locatio, '?'))
    {
      unsigned char *modified_url;
      locatio = base_uri (( char *)locatio, f->r, (unsigned char *)locatio);
        modified_url = apr_pstrndup ( f->r->pool, locatio  ,  strlen(locatio));
        if(   fctx->cfg->url_useip)
           modified_url =apr_pstrcat(f->r->pool, modified_url ,f->r->connection->remote_ip,NULL);
 
      unsigned char *hmac =
	fctx->cfg->url_funchandle (f->r, modified_url, strlen ( modified_url),
			       fctx->cfg->atpasswd,  ATPASSWDLEN );
      ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, f->r,
		     "LOCATIONE!!!  %s HMAC %s SERVER %s HOST %s ADDR %s",
		     locatio, hmac, f->r->server->server_hostname,
		     apr_table_get (f->r->headers_in, "Host"),
		     f->r->connection->local_ip);
      locatio =
	apr_pstrcat (f->r->pool, locatio, QUERY_SEPARATOR,
		     fctx->cfg->hmacprefix, hmac, NULL);
      apr_table_set (f->r->headers_out, "Location", locatio);
      //free(hmac);
    }


  if (f->r->proto_num >= 1001)
    {
      if (!f->r->main && !f->r->prev)
	{
	  env = apr_table_get (f->r->subprocess_env, "force-response-1.0");
	  if (!env)
	    f->r->chunked = 1;
	}
    }
  apr_table_unset (f->r->headers_out, "Content-Length");
  apr_table_unset (f->r->headers_out, "ETag");
  return OK;
}


static saxctxt *
check_filter_init (ap_filter_t * f)
{

  const char *errmsg = NULL;
  /*  if ( ! f->r->proxyreq ) {
     errmsg = "Non-proxy request; not inserting proxy-html filter" ;
     } else  */ if (!f->r->content_type)
    {
      errmsg = "No content-type; bailing out of proxy-html filter";
    }
  else if (strncasecmp (f->r->content_type, "text/html", 9) &&
	   strncasecmp (f->r->content_type, "application/xhtml+xml", 21))
    {
      errmsg = "Non-HTML content; not inserting proxy-html filter";
    }

  if (errmsg)
    {
#ifndef GO_FASTER
      anti_tamper_conf *cfg
	= ap_get_module_config (f->r->per_dir_config, &anti_tamper_module);
      if (cfg->verbose)
	{
	  ap_log_rerror (APLOG_MARK, APLOG_INFO, 0, f->r, errmsg);
	}
#endif

      ap_remove_output_filter (f);
      return NULL;
    }
  if (!f->ctx)
    anti_tamper_filter_init (f);
  return f->ctx;
}


static int
hmac_cookies (void *f, const char *key, const char *value)
{
  int clen;
  char *sc;
  ap_filter_t *filt = (ap_filter_t *) f;
  request_rec *r = (request_rec *) (filt->r);
  char *hmac;
  saxctxt *ctx = filt->ctx;
  char *hmacprefix =
    apr_pstrcat (r->pool, COOKIE_PREFIX, ctx->cfg->hmacprefix, NULL);
  apr_table_t *ct = ctx->cookie_table;
  char *cookie_val;
  char * modified_cookie;
  
  if (key == NULL || value == NULL || value[0] == '\0')
    return 1;
  if ((sc = strchr ((char *)value, ';')))
    clen = sc - value;
  else
    clen = strlen (value);
  cookie_val = apr_pstrndup (r->pool, value, clen);
  if (ctx->got_headers == 1 || strstr (cookie_val, hmacprefix))
    {
      return 1;
    }
    
   modified_cookie = apr_pstrndup (r->pool, cookie_val, clen);
  if(  ctx->cfg->cookie_useip)
    modified_cookie =apr_pstrcat(r->pool,modified_cookie,r->connection->remote_ip,NULL);
  
  hmac = ctx->cfg->cookie_funchandle (r, modified_cookie , strlen(modified_cookie), ctx->cfg->atpasswd, ATPASSWDLEN);

  cookie_val = apr_pstrcat (r->pool, cookie_val, hmacprefix, NULL);
  cookie_val = apr_pstrcat (r->pool, cookie_val, hmac, NULL);
  if (sc)
    cookie_val = apr_pstrcat (r->pool, cookie_val, sc, NULL);
  apr_table_add (ct, "Set-Cookie", cookie_val);
  return 1;
}

static int
add_cookies (void *f, const char *key, const char *value)
{
  apr_table_add (f, key, value);
#ifndef GO_FASTER
  fprintf (stderr, "Auth COOKIE add_cookies %s \n", value);
  fflush (stderr);
#endif
  return 1;
}

static int
anti_tamper_filter (ap_filter_t * f, apr_bucket_brigade * bb)
{
  apr_bucket *b;
  meta *m = NULL;
  xmlCharEncoding enc;
  const char *buf = 0;
  apr_size_t bytes = 0;
  int xmlopts = XML_PARSE_RECOVER | XML_PARSE_NONET |
    XML_PARSE_NOBLANKS | XML_PARSE_NOERROR | XML_PARSE_NOWARNING;

  saxctxt *ctxt = check_filter_init (f);
  if (!ctxt)
    return ap_pass_brigade (f->next, bb);

  ctxt->cookie_table = apr_table_make (f->r->pool, 10);
  if (ctxt->cfg->wants_cookie_anti_tamper)
    {
      apr_table_do (hmac_cookies, f, f->r->headers_out, "Set-Cookie", NULL);
      // ap_log_rerror(APLOG_MARK, APLOG_ERR, NULL, r,
      //                           "apache2_ext_filter: Response.%s \n", cookie1 );
      apr_table_unset (f->r->headers_out, "Set-Cookie");
      apr_table_do (add_cookies, f->r->headers_out, ctxt->cookie_table,
		    "Set-Cookie", NULL);
      ctxt->got_headers = 1;
    }
  // ctxt->host=;
  for (b = APR_BRIGADE_FIRST (bb);
       b != APR_BRIGADE_SENTINEL (bb); b = APR_BUCKET_NEXT (b))
    {
      if (APR_BUCKET_IS_EOS (b))
	{
	  if (ctxt->parser != NULL)
	    {
	      htmlParseChunk (ctxt->parser, buf, 0, 1);
	    }
	  APR_BRIGADE_INSERT_TAIL (ctxt->bb,
				   apr_bucket_eos_create (ctxt->bb->
							  bucket_alloc));
	  ap_pass_brigade (ctxt->f->next, ctxt->bb);
	}
      else if (apr_bucket_read (b, &buf, &bytes, APR_BLOCK_READ)
	       == APR_SUCCESS)
	{
	  if (ctxt->parser == NULL)
	    {
	      if (buf[bytes] != 0)
		{
		  /* make a string for parse routines to play with */
		  char *buf1 = apr_palloc (f->r->pool, bytes + 1);
		  memcpy (buf1, buf, bytes);
		  buf1[bytes] = 0;
		  buf = buf1;
		}
#ifndef GO_FASTER
	      enc = sniff_encoding (f->r, buf, bytes, ctxt->cfg->verbose);
	      if (ctxt->cfg->metafix)
		m = metafix (f->r, buf, ctxt->cfg->verbose);
#else
	      enc = sniff_encoding (f->r, buf, bytes);
	      if (ctxt->cfg->metafix)
		m = metafix (f->r, buf);
#endif
	      ap_set_content_type (f->r, "text/html;charset=utf-8");
	      ap_fputs (f->next, ctxt->bb, ctxt->cfg->doctype);
	      if (m)
		{
		  ctxt->parser = htmlCreatePushParserCtxt (ctxt->sax, ctxt,
							   buf, m->start, 0,
							   enc);
		  htmlParseChunk (ctxt->parser, buf + m->end, bytes - m->end,
				  0);
		}
	      else
		{
		  ctxt->parser = htmlCreatePushParserCtxt (ctxt->sax, ctxt,
							   buf, bytes, 0,
							   enc);
		}
	      apr_pool_cleanup_register (f->r->pool, ctxt->parser,
					 (void *) htmlFreeParserCtxt,
					 apr_pool_cleanup_null);
	      if (xmlopts =
		  xmlCtxtUseOptions (ctxt->parser, xmlopts), xmlopts)
		ap_log_rerror (APLOG_MARK, APLOG_WARNING, 0, f->r,
			       "Unsupported parser opts %x", xmlopts);
	    }
	  else
	    {
	      htmlParseChunk (ctxt->parser, buf, bytes, 0);
	    }
	}
      else
	{
	  ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, f->r,
			 "Error in bucket read");
	}
    }

  /*ap_fflush(ctxt->f->next, ctxt->bb) ;        // uncomment for debug */
  apr_brigade_cleanup (bb);
  return APR_SUCCESS;
}
static const char *fpi_html =
  "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n";
static const char *fpi_html_legacy =
  "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n";
static const char *fpi_xhtml =
  "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n";
static const char *fpi_xhtml_legacy =
  "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n";
static const char *html_etag = ">";
static const char *xhtml_etag = " />";
/*#define DEFAULT_DOCTYPE fpi_html */
static const char *DEFAULT_DOCTYPE = "";
#define DEFAULT_ETAG html_etag

static void *
anti_tamper_config(apr_pool_t * pool, char *x)
{
  anti_tamper_conf *ret =(anti_tamper_conf *)apr_pcalloc (pool, sizeof (anti_tamper_conf));
    if (ret == NULL) return NULL;

  ret->doctype = DEFAULT_DOCTYPE;
  ret->etag = DEFAULT_ETAG;
  ret->bufsz = 8192;
  ret->wants_cookie_anti_tamper = 1;
  ret->wants_url_anti_tamper = 0;
  ret->redir_url = apr_pstrdup (pool, "/");
  ret->hmacprefix = apr_pstrdup (pool, DEFAULT_AUTHVAR_PREFIX);
  ret->secret_path = apr_pstrdup (pool, SECRET_PATH);
  ret->dontmodifyprefix = apr_pstrdup (pool, DEFAULT_DONTMODIFYVAR_PREFIX);
  ret->url_funchandle = get_hmac_string;
  ret->cookie_funchandle = get_hmac_string;
  ret->url_hmacresponselen = DEFAULT_HMACRESP_LEN  ;
  ret->cookie_hmacresponselen = DEFAULT_HMACRESP_LEN  ;
  ret->exclude_args = apr_array_make(pool, 10, sizeof ( char *));
  
 if(!getuid()){
  int handle =
    open (ret->secret_path, O_CREAT | O_RDWR,
	  (S_IRUSR | S_IWUSR));
  if (handle == -1 )
  {
    fprintf (stderr, " errno=%d(%s)-%x\n%s", errno, strerror (errno),*(ret->atpasswd),x);
    fflush (stderr);
  }
  else
    {
      int b;
      if ((b = read (handle, ret->atpasswd, ATPASSWDLEN)) < ATPASSWDLEN)
	{
	  fprintf (stderr, " rigenero %d\n", b);
	  fflush (stderr);
	  apr_generate_random_bytes (ret->atpasswd, ATPASSWDLEN);
	  write (handle, ret->atpasswd, ATPASSWDLEN);
	}
      close (handle);
    }
  }
  return ret;
}

static void *
anti_tamper_merge (apr_pool_t * pool, void *BASE, void *ADD)
{
  anti_tamper_conf *base = (anti_tamper_conf *) BASE;
  anti_tamper_conf *add = (anti_tamper_conf *) ADD;
  anti_tamper_conf *conf = apr_palloc (pool, sizeof (anti_tamper_conf));

  memcpy (conf, base, sizeof (*base));

  if (add->map && base->map)
    {
      urlmap *a;
      conf->map = NULL;
      for (a = base->map; a; a = a->next)
	{
	  urlmap *save = conf->map;
	  conf->map = apr_pmemdup (pool, a, sizeof (urlmap));
	  conf->map->next = save;
	}
      for (a = add->map; a; a = a->next)
	{
	  urlmap *save = conf->map;
	  conf->map = apr_pmemdup (pool, a, sizeof (urlmap));
	  conf->map->next = save;
	}
    }
  else
    conf->map = add->map ? add->map : base->map;

  conf->doctype = (add->doctype == DEFAULT_DOCTYPE)
    ? base->doctype : add->doctype;
  conf->etag = (add->etag == DEFAULT_ETAG) ? base->etag : add->etag;
  conf->bufsz = add->bufsz;

  conf->wants_url_anti_tamper = add->wants_url_anti_tamper;
  conf->wants_cookie_anti_tamper = add->wants_cookie_anti_tamper;
  conf->cookie_useip = add->cookie_useip ;
  conf->url_useip    = add->url_useip;
  conf->redir_url    = add->redir_url;


  if (add->flags & NORM_RESET)
    {
      conf->flags = add->flags ^ NORM_RESET;
      conf->metafix = add->metafix;
      conf->extfix = add->extfix;
      conf->strip_comments = add->strip_comments;
      conf->wants_url_anti_tamper = add->wants_url_anti_tamper;
      conf->wants_cookie_anti_tamper = add->wants_cookie_anti_tamper;
#ifndef GO_FASTER
      conf->verbose = add->verbose;
#endif
    }
  else
    {
      conf->flags = base->flags | add->flags;
      conf->metafix = base->metafix | add->metafix;
      conf->extfix = base->extfix | add->extfix;
      conf->strip_comments = base->strip_comments | add->strip_comments;
#ifndef GO_FASTER
      conf->verbose = base->verbose | add->verbose;
#endif
    }
  return conf;
}

#define REGFLAG(n,s,c) ( (s&&(strchr((s),(c))!=NULL)) ? (n) : 0 )
#define XREGFLAG(n,s,c) ( (!s||(strchr((s),(c))==NULL)) ? (n) : 0 )
static const char *
set_urlmap (cmd_parms * cmd, void *CFG,
	    const char *from, const char *to, const char *flags)
{
  int regflags;
  anti_tamper_conf *cfg = (anti_tamper_conf *) CFG;
  urlmap *map;
  urlmap *newmap = apr_palloc (cmd->pool, sizeof (urlmap));

  newmap->next = NULL;
  newmap->flags
    = XREGFLAG (M_HTML, (char *)flags, 'h')
    | XREGFLAG (M_EVENTS, (char *)flags, 'e')
    | XREGFLAG (M_CDATA,(char *) flags, 'c')
    | REGFLAG (M_ATSTART, (char *)flags, '^')
    | REGFLAG (M_ATEND, (char *)flags, '$')
    | REGFLAG (M_REGEX, (char *)flags, 'R') | REGFLAG (M_LAST, (char *)flags, 'L');

  if (cfg->map)
    {
      for (map = cfg->map; map->next; map = map->next);
      map->next = newmap;
    }
  else
    cfg->map = newmap;

  if (!(newmap->flags & M_REGEX))
    {
      newmap->from.c = apr_pstrdup (cmd->pool, from);
      newmap->to = apr_pstrdup (cmd->pool, to);
    }
  else
    {
      regflags
	= REGFLAG (REG_EXTENDED,(char *) flags, 'x')
	| REGFLAG (REG_ICASE,(char *) flags, 'i')
	| REGFLAG (REG_NOSUB,(char *) flags, 'n') | REGFLAG (REG_NEWLINE, (char *)flags, 's');
      newmap->from.r = ap_pregcomp (cmd->pool, from, regflags);
      newmap->to = apr_pstrdup (cmd->pool, to);
    }
  return NULL;
}
static const char *
set_doctype (cmd_parms * cmd, void *CFG, const char *t, const char *l)
{
  anti_tamper_conf *cfg = (anti_tamper_conf *) CFG;
  if (!strcasecmp (t, "xhtml"))
    {
      cfg->etag = xhtml_etag;
      if (l && !strcasecmp (l, "legacy"))
	cfg->doctype = fpi_xhtml_legacy;
      else
	cfg->doctype = fpi_xhtml;
    }
  else if (!strcasecmp (t, "html"))
    {
      cfg->etag = html_etag;
      if (l && !strcasecmp (l, "legacy"))
	cfg->doctype = fpi_html_legacy;
      else
	cfg->doctype = fpi_html;
    }
  else
    {
      cfg->doctype = apr_pstrdup (cmd->pool, t);
      if (l && ((l[0] == 'x') || (l[0] == 'X')))
	cfg->etag = xhtml_etag;
      else
	cfg->etag = html_etag;
    }
  return NULL;
}

static void
set_param (anti_tamper_conf * cfg, const char *arg)
{
  if (arg && *arg)
    {
      if (!strcmp (arg, "lowercase"))
	cfg->flags |= NORM_LC;
      else if (!strcmp (arg, "dospath"))
	cfg->flags |= NORM_MSSLASH;
      else if (!strcmp (arg, "reset"))
	cfg->flags |= NORM_RESET;
    }
}


static const char *
set_flags (cmd_parms * cmd, void *CFG, const char *arg1,
	   const char *arg2, const char *arg3)
{
  set_param ((anti_tamper_conf *) CFG, arg1);
  set_param ((anti_tamper_conf *) CFG, arg2);
  set_param ((anti_tamper_conf *) CFG, arg3);
  return NULL;
}

static const char *
set_hmac_url_function (cmd_parms * cmd, void *in_dcfg, const char *libname,
		   const char *funcname,const  char *len)
{
  void *handle;

  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;

  handle = dlopen (libname, RTLD_LAZY | RTLD_GLOBAL);
  if (!handle)
    {
      return apr_pstrcat (cmd->pool, "Cannot load ", libname,
			  " into server: ", dlerror (), NULL);
    }
  dcfg->url_funchandle =
    (char *(*)(request_rec *, char *, int, const unsigned char *, int))
    dlsym (handle, funcname);
//dcfg->funchandle=NULL;
  if (!dcfg->url_funchandle)
    {
      dlclose (handle);
      return apr_pstrcat (cmd->pool, "Cannot load from ", libname,
			  " Function ", funcname, " into server: ",
			  dlerror (), NULL);
    }
  dcfg->url_hmacresponselen= strtol( len, (char **)NULL, 10);
  return NULL;
}

static const char *
set_hmac_cookie_function (cmd_parms * cmd, void *in_dcfg, const char *libname,
		   const char *funcname,const  char *len)
{
  void *handle;

  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;


  handle = dlopen ( libname, RTLD_LAZY | RTLD_GLOBAL);
  if (!handle)
    {
      return apr_pstrcat (cmd->pool, "Cannot load ", libname,
			  " into server: ", dlerror (), NULL);
    }
  dcfg->cookie_funchandle =
    (char *(*)(request_rec *, char *, int, const unsigned char *, int))
    dlsym (handle, funcname);
//dcfg->funchandle=NULL;
  if (!dcfg->cookie_funchandle)
    {
      dlclose (handle);
      return apr_pstrcat (cmd->pool, "Cannot load from ", libname,
			  " Function ", funcname, " into server: ",
			  dlerror (), NULL);
    }
  dcfg->cookie_hmacresponselen= strtol( len, (char **)NULL, 10);
  return NULL;
}

/*todo*/
static const char *
add_args_to_exclude (cmd_parms * cmd, void *in_dcfg, const char *method,
		     const char *args)
{
    char *saveptr;
    char *t = apr_pstrdup(cmd->pool,args);
    char *p = strtok_r(t  , "|", &saveptr);
    anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;
    dcfg->exclude_args = apr_array_make(cmd->pool, 10, sizeof (char *));
 //    if(!strcmp(method,"GET"))
//     else if(!strcmp(method,"POST"))
//     else if(!strcmp(method,"BOTH"))
//     else
//     return apr_pstrcat (cmd->pool, "ERROR: invalid Method %s , use POST|GET|BOTH as values ",method, NULL); 
//     

    while (p != NULL) {
        char *arg = (char *)apr_pstrdup(cmd->pool, p);
        *(char **)apr_array_push(dcfg->exclude_args) =  arg;
       
        p = strtok_r(NULL, "|", &saveptr);
    }
    return NULL;
}



static const char *
set_redir_url (cmd_parms * cmd, void *in_dcfg, const char *url)
{
  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;

  dcfg->redir_url = (char *) url;

  return NULL;
}


static const char *
set_hmac_prefix (cmd_parms * cmd, void *in_dcfg, const char *prefix)
{
  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;

  dcfg->hmacprefix = apr_pstrcat (cmd->pool, prefix, "=", NULL);

  return NULL;
}
static const char *
set_secret_path (cmd_parms * cmd, void *in_dcfg, const char *path)
{
  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;

  dcfg->secret_path = apr_pstrdup (cmd->pool, path);

  return NULL;
}


static const char *
set_dontmodify_prefix (cmd_parms * cmd, void *in_dcfg, const char *prefix)
{
  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;

  dcfg->dontmodifyprefix = apr_pstrdup (cmd->pool, prefix);

  return NULL;
}
static const char *
set_hmac_uses_ip (cmd_parms * cmd, void *in_dcfg, const char *useipfor)
{
  anti_tamper_conf *dcfg = (anti_tamper_conf *) in_dcfg;
 
  if(strcmp(useipfor,"BOTH"))
 {
  dcfg->cookie_useip = 1;
  dcfg->url_useip = 1;
 }
  else if(strcmp(useipfor,"URL"))
  dcfg->url_useip = 1;
   else if(strcmp(useipfor,"COOKIE"))
  dcfg->cookie_useip = 1;
  else  if(strcmp(useipfor,"NONE"))
  {
   dcfg->cookie_useip = 0;
   dcfg->url_useip = 0;
  }else
  return apr_pstrcat (cmd->pool, "ATHmacUseIP: invalid Option %s , use URL|COOKIE|BOTH|NONE as values ",useipfor, NULL); 
  return NULL;
}

static const command_rec anti_tamper_cmds[] = {
  AP_INIT_TAKE23 ("ProxyHTMLURLMap", set_urlmap, NULL,
		  RSRC_CONF | ACCESS_CONF, "Map URL From To"),
  AP_INIT_TAKE12 ("ProxyHTMLDoctype", set_doctype, NULL,
		  RSRC_CONF | ACCESS_CONF, "(HTML|XHTML) [Legacy]"),
  AP_INIT_TAKE123 ("ProxyHTMLFixups", set_flags, NULL,
		   RSRC_CONF | ACCESS_CONF, "Options are lowercase, dospath"),
  AP_INIT_FLAG ("ProxyHTMLMeta", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, metafix),
		RSRC_CONF | ACCESS_CONF, "Fix META http-equiv elements"),
  AP_INIT_FLAG ("ProxyHTMLExtended", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, extfix),
		RSRC_CONF | ACCESS_CONF, "Map URLs in Javascript and CSS"),
  AP_INIT_FLAG ("ProxyHTMLStripComments", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, strip_comments),
		RSRC_CONF | ACCESS_CONF, "Strip out comments"),
#ifndef GO_FASTER
  AP_INIT_FLAG ("ProxyHTMLLogVerbose", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, verbose),
		RSRC_CONF | ACCESS_CONF,
		"Verbose Logging (use with LogLevel Info)"),
#endif
  AP_INIT_TAKE1 ("ProxyHTMLBufSize", ap_set_int_slot,
		 (void *) APR_OFFSETOF (anti_tamper_conf, bufsz),
		 RSRC_CONF | ACCESS_CONF, "Buffer size"),

  AP_INIT_FLAG ("ATHmacUrl", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, wants_url_anti_tamper),
		RSRC_CONF | OR_AUTHCFG, "ATHmacUrl [On|Off] --Set url anti anti_tamper filter ON "),
  AP_INIT_FLAG ("ATHmacCookie", ap_set_flag_slot,
		(void *) APR_OFFSETOF (anti_tamper_conf, wants_cookie_anti_tamper),
		RSRC_CONF | OR_AUTHCFG, "ATHmacCookie [On|Off] -- Set cookie anti anti_tamper filter ON "),
  AP_INIT_TAKE1 ("ATHmacUseIP", set_hmac_uses_ip ,
		 NULL,
		 RSRC_CONF | OR_AUTHCFG,  
		 "ATHmacUseIP [URL|COOKIE|BOTH|NONE] -- ,Add source IP to HMAC computation for respectively url,cookie,both or none"  ),
  AP_INIT_TAKE3 ("ATSetHmacUrlLibrary",
		 set_hmac_url_function,
		 NULL,
		 RSRC_CONF,
		 "ATSetHmacUrlLibrary <full_path_solib> <exported_function> <authlen> -- function takes aguments char *exported_function ( request_rec *r , char *data,int data_len, const unsigned char *key,int key_len) "),
  AP_INIT_TAKE3 ("ATSetHmacCookieLibrary",
		 set_hmac_cookie_function,
		 NULL,
		 RSRC_CONF,
		 "ATSetHmacCookieLibrary <full_path_solib> <exported_function> <authlen> -- function takes aguments char *exported_function ( request_rec *r , char *data,int data_len, const unsigned char *key,int key_len) "),
  AP_INIT_TAKE2 ("ATExcludeArgsForUri",
		 add_args_to_exclude,
		 NULL,
		 RSRC_CONF,
		 "ATExcludeArgsForUri [GET|POST|BOTH] arg1|arg2... -- The arguments listed here (and their value) will be excluded when computing the mac. Todo... Works only for GET"),
  AP_INIT_TAKE1 ("ATRedirect_url",
		 set_redir_url,
		 NULL,
		 RSRC_CONF | OR_AUTHCFG,
		 "ATRedirect_url <url>"),
  AP_INIT_TAKE1 ("ATSetHmacPrefix",
		 set_hmac_prefix,
		 NULL,
		 RSRC_CONF,
		 "ATSetHmacPrefix <prefix> -- Default is _auth but it could be changed (use A-Za-z0-9 subset)."),
  AP_INIT_TAKE1 ("ATSetSecretPath",
		 set_secret_path,
		 NULL,
		 RSRC_CONF,
		 "ATSetSecretPath <fullpath> -- Default is /etc/httpd/conf/anti_tamper_secret"),
  AP_INIT_TAKE1 ("ATSetDontModifyPrefix",
		 set_dontmodify_prefix,
		 NULL,
		 RSRC_CONF,
		 "ATSetDontModifyPrefix <prefix> -- Default is 'dontmodify' but it could be changed (use A-Za-z0-9 subset).See Documentation."),
  {NULL}
};

static int
mod_anti_tamper (apr_pool_t * p, apr_pool_t * p1, apr_pool_t * p2, server_rec * s)
{
  ap_add_version_component (p, VERSION_STRING);
  return OK;
}



char *
is_cookie_ok (const char *cookie, request_rec * r, anti_tamper_conf * dcfg)
{
  char *header_copy,  *p;
  int cookie_len = 0;
  char *real_cookie = NULL;
  char *one_real_cookie = NULL;
  real_cookie = apr_pstrndup (r->pool, "", 0);
  char *hmacprefix =
    apr_pstrcat (r->pool, COOKIE_PREFIX, dcfg->hmacprefix, NULL);
  int hmacprefixlen = strlen (hmacprefix);
  header_copy = apr_pstrdup (r->pool, cookie);
  if (header_copy == NULL)
    return NULL;

  apr_table_unset (r->headers_in, "Cookie");


  p = header_copy;
  cookie_len = strlen (p);
  while (*p != 0)
    {
      char *sc, *c, *auc;
      //find next semi colon
      if (!(sc = strchr (p, ';')))
	sc = p + cookie_len;

      c = apr_pstrndup (r->pool, p, sc - p);

      if ((auc = strstr (c, hmacprefix)))
	{

	  char *aut_val, *modified_cookie; 
	  /*Must be the last one or nothing */
	  aut_val = apr_pstrndup (r->pool, auc + hmacprefixlen, dcfg->cookie_hmacresponselen );
	  /* from auc to sc is the auth part; from p to auc is the real cookie part */
	  one_real_cookie = apr_pstrndup (r->pool, p, auc - c);
	  apr_collapse_spaces (one_real_cookie, one_real_cookie);
          
	  modified_cookie = apr_pstrdup (r->pool,  one_real_cookie);
	  if(dcfg->cookie_useip)
	    modified_cookie = apr_pstrcat(r->pool,modified_cookie, r->connection->remote_ip,NULL);
	  unsigned char *hmac =
	    dcfg->cookie_funchandle (r, modified_cookie , strlen (modified_cookie ),
			      dcfg->atpasswd, ATPASSWDLEN);


	  if (!strcmp (hmac, aut_val))
	    {
	      real_cookie =
		apr_pstrcat (r->pool, real_cookie, one_real_cookie, NULL);
#ifndef GO_FASTER
	      fprintf (stderr, "anti_tamper: Cookie Allowed!! %s\n",
		       one_real_cookie);
#endif

	    }
	  else
	    {
#ifndef GO_FASTER
	      fprintf (stderr, "anti_tamper: Cookie Not Allowed!! %s\n",
		       one_real_cookie);
#endif

	    }
	  if (!strcmp (hmac, aut_val) && sc != p + cookie_len)
	    real_cookie = apr_pstrcat (r->pool, real_cookie, ";", NULL);
	  fflush (stderr);
	}
      else
	{
#ifndef GO_FASTER
	  fprintf (stderr, "anti_tamper: Cookie Not Allowed!! %s\n", c);
#endif

	}


      if (sc != p + cookie_len)
	p = sc + 1;
      else
	break;
    }
  if (real_cookie != NULL)
    apr_table_set (r->headers_in, "Cookie", real_cookie);

  return NULL;
}

static char *exclude_args_from_uri(request_rec * r,anti_tamper_conf * dcfg )
{
   char *uri = apr_pstrdup (r->pool, r->unparsed_uri);
   char *p;
   char *t;
   char *saveptr;
   char *query;
   char *newquery;
   char **args;
   int i;
   query=strchr(uri,'?');
  
   p= apr_pstrdup( r->pool, query+1);
   newquery=(char *)apr_pcalloc( r->pool,sizeof(char ));
   *newquery='\0';
   t=strtok_r(p,"&",&saveptr);
   
   args = (char **)dcfg->exclude_args->elts;
   if(dcfg->exclude_args->nelts)
   {
    while(t!=NULL)
   {
    char *val=strchr(t,'=');
        for (i = 0; i <  dcfg->exclude_args->nelts; i++) 
    	 if(!strncmp(t, args[i],val-t))
      	   break;
         else
    	   newquery=apr_pstrcat( r->pool,newquery,(*newquery?"&":""),t,NULL);

     t = strtok_r(NULL,"&",&saveptr);
   }
   }else
    return  apr_pstrcat( r->pool,r->parsed_uri.path,query,NULL);
   
		 
   return  apr_pstrcat( r->pool,r->parsed_uri.path,"?",newquery,NULL);
}

static int
is_request_ok (request_rec * r, anti_tamper_conf * dcfg)
{
  char *uri;
  char *hmacprefix =
    apr_pstrcat (r->pool, QUERY_SEPARATOR, dcfg->hmacprefix, NULL);
  char *auth = strstr (r->parsed_uri.query, hmacprefix);
  int hmacprefixlen = strlen (hmacprefix);

  if (auth)
    {
      if ( r->unparsed_uri != NULL && (uri=exclude_args_from_uri(r,dcfg )))
	{
	  char *auth_ptr;
	  if ((auth_ptr = strstr (uri, hmacprefix))) {
	      char *auth_val;
	      char *uri_wo_auth;
	      int auth_len;
	      char *end_of_auth;
              char *modified_url;

	      for (end_of_auth = auth_ptr;
		   *end_of_auth == '&' || *end_of_auth == '\0';
		   end_of_auth++);
	      auth_val =
		apr_pstrndup (r->pool, auth_ptr + hmacprefixlen,
			      end_of_auth - auth_ptr - hmacprefixlen);
	      auth_len = strlen (auth_ptr);


	      if (auth_len < (hmacprefixlen + dcfg->url_hmacresponselen ))
		return 0;

	      uri_wo_auth = apr_pstrndup (r->pool, uri, auth_ptr - uri);

	      if (auth_len > (hmacprefixlen + dcfg->url_hmacresponselen))	// is in the middle of query, lets append the remaining string to uri_wo_auth
		uri_wo_auth =
		  apr_pstrcat (r->pool, uri_wo_auth,
			       auth_ptr + dcfg->url_hmacresponselen + hmacprefixlen, NULL);
               
	      modified_url = apr_pstrndup (r->pool,uri_wo_auth  , strlen (uri_wo_auth) );
	      if(  dcfg->url_useip)
		modified_url =apr_pstrcat(r->pool,modified_url,r->connection->remote_ip,NULL);
 
	      unsigned char *hmac =
		dcfg->url_funchandle (r, modified_url , strlen (modified_url  ),
				  dcfg->atpasswd, ATPASSWDLEN);
	      if (!strcmp (hmac, auth_val)) {
#ifndef GO_FASTER
		  fprintf (stderr,
			   "Auth_uri Accepted!.%s .%s %s %s %s %s %d\n",
			   r->args, r->uri, uri, auth_val, hmac, uri_wo_auth,
			   ap_is_initial_req (r));
#endif

		  if (r->args)
		    r->args =
		      apr_pstrdup (r->pool, strchr (uri_wo_auth, '?') + 1);

		  return 1;
		} else {

#ifndef GO_FASTER
		  fprintf (stderr, "Auth uri: Not Accepted!! %s %s\n", uri,
			   uri_wo_auth);
#endif
		 
		  return 0;
		}
	    }
	}
      return 1;
    }
  else
    return 0;
}


static int
mod_anti_tamper_method_handler (request_rec * r)
{

  const char *cookies = apr_table_get (r->headers_in, "Cookie");
  char *realcookies;



  anti_tamper_conf *dcfg =
    (anti_tamper_conf *) ap_get_module_config (r->per_dir_config, &anti_tamper_module);


  if (dcfg->wants_cookie_anti_tamper
      && (realcookies = is_cookie_ok (cookies, r, dcfg)))
    {
      apr_table_setn (r->headers_in, "Cookie", realcookies);
    }


  if (dcfg->wants_url_anti_tamper && r->parsed_uri.query != NULL
      && !(is_request_ok (r, dcfg)))
    {
#ifndef GO_FASTER
      fprintf (stderr, "anti_tamper_Output: Not Allowed! %s %s\n", cookies,
	       (r->parsed_uri).query);
      fflush (stderr);
#endif

      apr_table_setn (r->headers_out, "Location", dcfg->redir_url);
      r->status = HTTP_MOVED_TEMPORARILY;
      return HTTP_MOVED_TEMPORARILY;
    }


  // Return DECLINED so that the Apache core will keep looking for
  // other modules to handle this request.  This effectively makes
  // this module completely transparent.
  return DECLINED;
}
static void at_pre(request_rec *r) {
     anti_tamper_conf *dcfg =
        (anti_tamper_conf *) ap_get_module_config (r->per_dir_config, &anti_tamper_module);
    
    if (dcfg == NULL) return;
   
   
    ap_add_output_filter_handle(global_at_filter_out, NULL, r, r->connection);
    
    
}
static void
anti_tamper_hooks (apr_pool_t * p)
{
  ap_hook_handler (mod_anti_tamper_method_handler, NULL, NULL, APR_HOOK_FIRST);
  global_at_filter_out=ap_register_output_filter ("anti_tamper_filter", anti_tamper_filter,
			     NULL, AP_FTYPE_RESOURCE);
  ap_hook_post_config (mod_anti_tamper, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_insert_filter(at_pre, NULL, NULL, APR_HOOK_FIRST);
  ap_hook_child_init (anti_tamper_child_init, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA anti_tamper_module = {
  STANDARD20_MODULE_STUFF,
  anti_tamper_config,
  anti_tamper_merge,
  NULL,
  NULL,
  anti_tamper_cmds,
  anti_tamper_hooks
};
