/* apache */
#include <http_protocol.h>
#include <http_config.h>
#include <http_log.h>
#include <apr_strings.h>
#include <apr_sha1.h>

char *get_hmac_string( request_rec *r, char *msg,int msglen, const unsigned char *key,int key_len)
{
    apr_sha1_ctx_t ctx;
    unsigned char digest[APR_SHA1_DIGESTSIZE];
    unsigned char k_ipad[65], k_opad[65];
    unsigned char kt[APR_SHA1_DIGESTSIZE];
    unsigned char *k = (unsigned char *)key;
    char hex_digest[APR_SHA1_DIGESTSIZE * 2], *hd;
    const char hex[] = "0123456789abcdef";
    int   i;
    apr_pool_t *p=r->pool;
    
     /* if key is longer than 64 bytes reset it to key=SHA1(key) */
    if (key_len > 64) {
        k = kt;
        key_len = APR_SHA1_DIGESTSIZE; 
    }

    /* HMAC_MD5 transform */
    /* start out by string key in pads */
    memset((void *)k_ipad, 0, sizeof(k_ipad));
    memset((void *)k_opad, 0, sizeof(k_opad));
    memmove(k_ipad, k, key_len);
    memmove(k_opad, k, key_len);

    /* XOR key with ipad and opad values */
    for (i = 0; i < 64; i++) {
        k_ipad[i] ^= 0x36;
        k_opad[i] ^= 0x5c;
    }
    /* perform inner SHA-1 */
     apr_sha1_init(&ctx);
     apr_sha1_update_binary(&ctx, k_ipad, 64);
    apr_sha1_update_binary(&ctx, (const unsigned char *)msg, msglen);
    apr_sha1_final(digest, &ctx);
    /* perform outer SHA-1 */
    apr_sha1_init(&ctx);
    apr_sha1_update_binary(&ctx, k_opad, 64);
    apr_sha1_update_binary(&ctx, digest, sizeof(digest));
     apr_sha1_final(digest, &ctx);

    hd = hex_digest;
    for (i = 0; i < sizeof(digest); i++) {
        *hd++ = hex[digest[i] >> 4];
        *hd++ = hex[digest[i] & 0xF];
    }
    *hd = '\0';

    return apr_pstrdup(p, hex_digest);
}
